
#ifndef PLS_TBB_LIKE_TASK_H
#define PLS_TBB_LIKE_TASK_H

#include "pls/internal/helpers/profiler.h"

#include "pls/internal/data_structures/aligned_stack.h"
#include "pls/internal/data_structures/deque.h"

#include "abstract_task.h"
#include "thread_state.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            class fork_join_task;
            class fork_join_sub_task: public data_structures::deque_item {
                friend class fork_join_task;

                // Coordinate finishing of sub_tasks
                std::atomic_uint32_t ref_count_;
                fork_join_sub_task* parent_;

                // Access to TBB scheduling environment
                fork_join_task* tbb_task_;

                // Stack Management (reset stack pointer after wait_for_all() calls)
                data_structures::aligned_stack::state stack_state_;
            protected:
                explicit fork_join_sub_task();
                fork_join_sub_task(const fork_join_sub_task& other);

                // Overwritten with behaviour of child tasks
                virtual void execute_internal() = 0;

            public:
                // Only use them when actually executing this sub_task (only public for simpler API design)
                template<typename T>
                void spawn_child(const T& sub_task);
                void wait_for_all();

            private:
                void spawn_child_internal(fork_join_sub_task* sub_task);
                void execute();
            };

            template<typename Function>
            class fork_join_lambda: public fork_join_sub_task {
                const Function* function_;

            public:
                explicit fork_join_lambda(const Function* function): function_{function} {};

            protected:
                void execute_internal() override {
                    (*function_)(this);
                }
            };

            class fork_join_task: public abstract_task {
                friend class fork_join_sub_task;

                fork_join_sub_task* root_task_;
                fork_join_sub_task* currently_executing_;
                data_structures::aligned_stack* my_stack_;

                // Double-Ended Queue management
                data_structures::deque<fork_join_sub_task> deque_;

                // Steal Management
                fork_join_sub_task* last_stolen_;

                fork_join_sub_task* get_local_sub_task();
                fork_join_sub_task* get_stolen_sub_task();

                bool internal_stealing(abstract_task* other_task) override;
                bool split_task(base::spin_lock* /*lock*/) override;

            public:
                explicit fork_join_task(fork_join_sub_task* root_task, const abstract_task::id& id);
                void execute() override;
                fork_join_sub_task* currently_executing() const;
            };

            template<typename T>
            void fork_join_sub_task::spawn_child(const T& task) {
                PROFILE_FORK_JOIN_STEALING("spawn_child")
                static_assert(std::is_base_of<fork_join_sub_task, T>::value, "Only pass fork_join_sub_task subclasses!");

                T* new_task = tbb_task_->my_stack_->push(task);
                spawn_child_internal(new_task);
            }
        }
    }
}

#endif //PLS_TBB_LIKE_TASK_H
